/*
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Charles P. Wright
 * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2004 Puja Gupta
 * Copyright (c) 2003-2004 Harikesavan Krishnan
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: file.c,v 1.68 2004/11/10 14:38:03 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"
#include <linux/dcache.h>

/*******************
 * File Operations *
 *******************/

static int unionfs_file_revalidate(struct file *file) {
    struct super_block *sb;
    struct dentry *dentry;
    int sbgen, fgen, dgen;
    int bindex, bstart, bend;
    struct file *hidden_file;
    struct dentry *hidden_dentry;

    int err = 0;

    print_entry(" file = %p", file);

    PASSERT(file);
    PASSERT(ftopd(file));

    dentry = file->f_dentry;
    PASSERT(dentry);
    PASSERT(dentry->d_op);
    PASSERT(dentry->d_op->d_revalidate);
    if (!(dentry->d_op->d_revalidate(dentry, 0))) {
	err = -ESTALE;
	goto out;
    }

    sb = dentry->d_sb;
    PASSERT(sb);
    sbgen = atomic_read(&stopd(sb)->usi_generation);
    dgen = atomic_read(&dtopd(dentry)->udi_generation);
    fgen = atomic_read(&ftopd(file)->ufi_generation);
    if (sbgen > dgen) {
	FISTBUG("The dentry is not up to date!\n");
    }
    if (sbgen > fgen) {
	/* First we throw out the existing files. */
	bstart = fbstart(file);
	bend = fbend(file);
	for (bindex = bstart; bindex <= bend; bindex++) {
	    if (ftohf_index(file, bindex)) {
		branchput(dentry->d_sb, bindex);
		fput(ftohf_index(file, bindex));
	    }
	}	
	if (ftohf_ptr(file)) {
	    KFREE(ftohf_ptr(file));
	}

	/* Now we reopen the file(s) as in unionfs_open. */
	bstart = fbstart(file) = dbstart(dentry);
	bend = fbend(file) = dbend(dentry);

	ftohf_ptr(file) = KMALLOC(sizeof(file_t *) * sbmax(sb), GFP_KERNEL);
	if (!ftohf_ptr(file)) {
	    err = -ENOMEM;
	    goto out;
	}

	init_ftohf_ptr(file, sbmax(sb));


	if (S_ISDIR(dentry->d_inode->i_mode)) {
	    /* We need to open all the files. */
	    for (bindex = bstart; bindex <= bend; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
		    continue;
		}

		dget(hidden_dentry);
		mntget(stohiddenmnt_index(sb, bindex));
		branchget(sb, bindex);

		hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), file->f_flags);
		if (IS_ERR(hidden_file)) {
		    err = PTR_ERR(hidden_file);
		    goto out;
		} else {
		    ftohf_index(file, bindex) = hidden_file;
		}
	    }
	} else {
	    /* We only open the highest priority branch. */
	    hidden_dentry = dtohd(dentry);
	    PASSERT(hidden_dentry);
	    PASSERT(hidden_dentry->d_inode);
	    if (IS_WRITE_FLAG(file->f_flags) && is_robranch(dentry)) {
		for (bindex = bstart - 1; bindex >= 0; bindex--) {
		    err = unionfs_copyup_file(dentry->d_parent->d_inode, file, bstart, bindex, file->f_dentry->d_inode->i_size);
		    if (!err) {
			break;
		    }
		}
		goto out;
	    }

	    dget(hidden_dentry);
	    mntget(stohiddenmnt_index(sb, bstart));
	    branchget(sb, bstart);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bstart), file->f_flags);
	    if (IS_ERR(hidden_file)) {
		err = PTR_ERR(hidden_file);
		goto out;
	    }
	    ftohf(file) = hidden_file;
	    /* Fix up the position. */
	    hidden_file->f_pos = file->f_pos;
	    /* update readahead information, all the time, because the old file
	     * could have read-ahead information that doesn't match.
             */
	    hidden_file->f_reada = file->f_reada;
	    hidden_file->f_ramax = file->f_ramax;
	    hidden_file->f_raend = file->f_raend;
	    hidden_file->f_ralen = file->f_ralen;
	    hidden_file->f_rawin = file->f_rawin;
	}
        atomic_set(&ftopd(file)->ufi_generation, atomic_read(&itopd(dentry->d_inode)->uii_generation));
    }

out:
    print_exit_status(err);
    return err;
}

STATIC loff_t
unionfs_llseek(file_t *file, loff_t offset, int origin)
{
    loff_t err;
    file_t *hidden_file = NULL;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    if (ftopd(file) != NULL) {
	hidden_file = ftohf(file);
    }

    fist_dprint(6, "unionfs_llseek: file=%p, offset=0x%llx, origin=%d\n",
		file, offset, origin);

    /* always set hidden position to this one */
    hidden_file->f_pos = file->f_pos;
    if (file->f_reada) { /* update readahead information if needed */
	hidden_file->f_reada = file->f_reada;
	hidden_file->f_ramax = file->f_ramax;
	hidden_file->f_raend = file->f_raend;
	hidden_file->f_ralen = file->f_ralen;
	hidden_file->f_rawin = file->f_rawin;
    }
    ASSERT(hidden_file->f_reada == file->f_reada);
    ASSERT(hidden_file->f_ramax == file->f_ramax);
    ASSERT(hidden_file->f_raend == file->f_raend);
    ASSERT(hidden_file->f_ralen == file->f_ralen);
    ASSERT(hidden_file->f_rawin == file->f_rawin);

    if (hidden_file->f_op && hidden_file->f_op->llseek)
	err = hidden_file->f_op->llseek(hidden_file, offset, origin);
    else
	err = generic_file_llseek(hidden_file, offset, origin);

    if (err < 0)
	goto out;
    if (err != file->f_pos) {
	file->f_pos = err;
	// ION maybe this?
	// 	file->f_pos = hidden_file->f_pos;
	file->f_reada = 0;
	file->f_version = ++event;
    }
out:
    print_exit_status(err);
    return err;
}


STATIC ssize_t
unionfs_read(file_t *file, char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    loff_t pos = *ppos;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    fist_print_file("entering read()", file);

    if (ftopd(file) != NULL) {
	hidden_file = ftohf(file);
    }
    
    if (!hidden_file->f_op || !hidden_file->f_op->read)
	goto out;

    err = hidden_file->f_op->read(hidden_file, buf, count, &pos);

    if (err >= 0) {
	/* atime should also be updated for reads of size zero or more */
	fist_copy_attr_atime(file->f_dentry->d_inode,
			     hidden_file->f_dentry->d_inode);
    }

    // MAJOR HACK
    /*
     * because pread() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    if (hidden_file->f_reada) { /* update readahead information if needed */
	file->f_reada = hidden_file->f_reada;
	file->f_ramax = hidden_file->f_ramax;
	file->f_raend = hidden_file->f_raend;
	file->f_ralen = hidden_file->f_ralen;
	file->f_rawin = hidden_file->f_rawin;
    }

out:
    print_exit_status(err);
    return err;
}


/* this unionfs_write() does not modify data pages! */
STATIC ssize_t
unionfs_write(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    inode_t *inode;
    inode_t *hidden_inode;
    loff_t pos = *ppos;
    int bstart, bend;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    inode = file->f_dentry->d_inode;

    bstart = fbstart(file);
    bend = fbend(file);

    ASSERT(bstart != -1);
    PASSERT(ftopd(file));
    PASSERT(ftohf(file));

    hidden_file = ftohf(file);      
    hidden_inode = hidden_file->f_dentry->d_inode;
    
    if (!hidden_file->f_op || !hidden_file->f_op->write) {
	goto out;
    }
    
    /* adjust for append -- seek to the end of the file */
    if (file->f_flags & O_APPEND)
	pos = inode->i_size;
	
    err = hidden_file->f_op->write(hidden_file, buf, count, &pos);
    
    /*
     * copy ctime and mtime from lower layer attributes
     * atime is unchanged for both layers
     */
    if (err >= 0)
	fist_copy_attr_times(inode, hidden_inode);
    
    /*
     * XXX: MAJOR HACK
     *
     * because pwrite() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on write() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    
    /* update this inode's size */
    if (pos > inode->i_size)
	inode->i_size = pos;

 out:
    print_exit_status(err);
    return err;
}

struct getdents_callback {
	struct linux_dirent * current_dir;
	struct linux_dirent * previous;
	int count;
	int error;
};

struct unionfs_getdents_callback {
    void *dirent;
    int flag_data_written;
    int filldir_called;
    file_t *this_file;
    filldir_t filldir;
    int bindex;
    int error;
};

/* copied from generic filldir in fs/readir.c */
STATIC int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
unionfs_filldir(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type)
# else /* kernel older than 2.4.9 */
unionfs_filldir(void *dirent, const char *name, int namelen, off_t offset, ino_t ino, unsigned int d_type)
# endif /* kernel older than 2.4.9 */
{
    struct unionfs_getdents_callback *buf = (struct unionfs_getdents_callback *) dirent;
    struct filldir_node *new = NULL;
    int err = 0;
    int found = 0;
    int index = 0;
    unsigned int hash = 0;
    int found_wh_entry = 0;

    /* For duplicate elimination. */
    struct filldir_node *cursor;
    struct list_head *pos;
    struct list_head *head;

   // fist_dprint(8, "unionfs_filldir name=%*s\n", namelen, name);

    buf->filldir_called++;
    
    if (!strncmp(name, ".wh.", 4)) {
	namelen = namelen - 4;
	hash = full_name_hash(&name[4], namelen);
	name = &name[4];
	found_wh_entry = 1;
    } else {
	hash = full_name_hash(name, namelen);
    }
    
    index = hash % HASHTABLE_SIZE;
    
    /* check to see if the entry exists already */
    head = &(ftopd(buf->this_file)->hash->dir_list[index]);

    /* See if we can find it in the list. */
    list_for_each(pos, head) {
	cursor = list_entry(pos, struct filldir_node, file_list);
	
	if (cursor->namelen == namelen && cursor->hash == hash && !strncmp(cursor->name, name, namelen)) {
	    /* a duplicate exists, and hence no need to create entry to the list */
	    found = 1;
	    /* if the duplicate is in this branch, then the file system is corrupted. */
	    if (cursor->bindex == buf->bindex) {
		//buf->error = err = -EIO;
		fist_dprint(8, "Possible I/O error unionfs_filldir: a file is duplicated in the same branch %d: %*s\n", buf->bindex, cursor->namelen, cursor->name);
		goto found_entry;
	    }
	    break;
	}
    }

    if (!found) {
	struct filldir_node *del;
	new = (struct filldir_node *)KMALLOC(sizeof(struct filldir_node), GFP_KERNEL);
        if (!new) {
            err = -ENOMEM;
            goto found_entry;
        }
	
	INIT_LIST_HEAD(&new->file_list);
	new->namelen = namelen;
	new->name = (char *) KMALLOC(namelen + 1, GFP_KERNEL);
	if (!new->name) {
	    buf->error = err = -ENOMEM;
	    goto found_entry;
	}
	new->hash = hash;
	new->bindex = buf->bindex;
	memcpy(new->name, name, namelen);
	new->name[namelen] = '\0';
	list_add(&(new->file_list), head);

	fist_dprint(8, "filldir added entry: %s\n", new->name);

	del = new;
	new = NULL; /* Since now it is linked in and we don't want it freed. */

	/* if 'name' was a whiteout entry just return without calling fill_dir */
	if (found_wh_entry) 
	  goto found_entry;

	buf->flag_data_written++;
	err = buf->filldir(buf->dirent, name, namelen, offset, ino, d_type);
	if (err) {
		buf->flag_data_written--;
		/* We need to take it out of the duplicate list. */
		fist_dprint(8, "filldir removed entry: %s\n", del->name);
		list_del(&del->file_list);		
		KFREE(del->name);
		KFREE(del);
	}
    }
    
 found_entry:
    if (new) {
	if (new->name) {
		KFREE(new->name);
	}
	KFREE(new);
    }
    return err;
}

STATIC int
unionfs_readdir(file_t *file, void *dirent, filldir_t filldir)
{
    int err = 0;
    file_t *hidden_file = NULL;
    inode_t *inode = NULL;
    struct unionfs_getdents_callback buf;
    int bstart, bend;
    int bindex = 0;

    print_entry_location();


    fist_print_file("In unionfs_readdir:", file);

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }
    
    inode = file->f_dentry->d_inode;
    fist_checkinode(inode, "unionfs_readdir");
   
    if(ftopd(file)->ufi_bindex >= 0) { 
    	bstart = ftopd(file)->ufi_bindex;
    } else {
    	bstart = fbstart(file);
    }
    fist_dprint(8, "Starting with branch %d\n", bstart);
    bend = fbend(file);

    bindex = bstart;
    while (bindex <= bend) {
	hidden_file = ftohf_index(file, bindex);
	if (!hidden_file) {
	    fist_dprint(8, "Incremented bindex to %d of %d because hidden file is NULL.\n", bindex, bend);
            bindex++;
	    continue;
	}
	hidden_file->f_pos = file->f_pos;
	    
	/* prepare callback buffer */
	buf.flag_data_written = 0;        /* means no data sent to user level */
	buf.filldir_called = 0;
	buf.dirent = dirent;
	buf.filldir = filldir;
	buf.this_file = file;
	buf.bindex = bindex;
	buf.error = 0;
   
	err = vfs_readdir(hidden_file, unionfs_filldir, (void *) &buf);
	if (buf.error) {
	    err = buf.error;
	    goto out;
	}
    
	/* set attributes */
	file->f_pos = hidden_file->f_pos;
	if (err > 0) {
	    fist_copy_attr_atime(inode, hidden_file->f_dentry->d_inode);
	}
	
	/* this means we read some data for this branch. */
	if (buf.flag_data_written) {
	    goto out;
	}
	if (!buf.filldir_called) {
		bindex++;
		file->f_pos = 0;
		fist_dprint(8, "Incremented bindex to %d of %d because filldir was not called.\n", bindex, bend);
	}
    }


    if (!err && bindex == bend) {
	free_hash(ftopd(file)->hash);
	ftopd(file)->ufi_bindex = -1;
    } else {
	ftopd(file)->ufi_bindex = bindex;
    }

 out:
    fist_checkinode(inode, "post unionfs_readdir");
    print_exit_status(err);
    return err;
}

STATIC int
unionfs_file_readdir(file_t *file, void *dirent, filldir_t filldir)
{
    int err = -ENOTDIR;
    print_entry_location();
    print_exit_status(err);
    return err;
}


STATIC unsigned int
unionfs_poll(file_t *file, poll_table *wait)
{
    unsigned int mask = DEFAULT_POLLMASK;
    file_t *hidden_file = NULL;

    print_entry_location();

    if (unionfs_file_revalidate(file)) {
	/* We should pretend an error happend. */
	mask = POLLERR|POLLIN|POLLOUT;
	goto out;
    }

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);

    if (!hidden_file->f_op || !hidden_file->f_op->poll)
	goto out;

    mask = hidden_file->f_op->poll(hidden_file, wait);

out:
    print_exit_status(mask);
    return mask;
}

STATIC int
unionfs_ioctl(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    int err = 0;		/* don't fail by default */
    file_t *hidden_file = NULL;
    vfs_t *this_vfs;
    vnode_t *this_vnode;
    int val;
#ifdef FIST_COUNT_WRITES
    extern unsigned long count_writes, count_writes_middle;
# endif /* FIST_COUNT_WRITES */

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    this_vfs = inode->i_sb;
    this_vnode = inode;

    /* check if asked for local commands */
    switch (cmd) {
    case FIST_IOCTL_GET_DEBUG_VALUE:
        if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
	val = fist_get_debug_value();
	err = put_user(val, (int *) arg);
	break;

    case FIST_IOCTL_SET_DEBUG_VALUE:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = get_user(val, (int *) arg);
	if (err)
	    break;
	fist_dprint(6, "IOCTL SET: got arg %d\n", val);
	if (val < 0 || val > 20) {
	    err = -EINVAL;
	    break;
	}
	fist_set_debug_value(val);
	break;

    /* add non-debugging fist ioctl's here */

    case UNIONFS_IOCTL_BRANCH_COUNT:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_branchcount(inode, file, cmd, arg);
	break;

    case UNIONFS_IOCTL_INCGEN:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_incgen(inode, file, cmd, arg);
	break;

    case UNIONFS_IOCTL_ADDBRANCH:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_addbranch(inode, file, cmd, arg);
	break;

    case UNIONFS_IOCTL_DELBRANCH:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_delbranch(inode, file, cmd, arg);
	break;

    case UNIONFS_IOCTL_RDWRBRANCH:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_rdwrbranch(inode, file, cmd, arg);
	break;

    case UNIONFS_IOCTL_SUPERDUPER:
	if (!capable(CAP_SYS_ADMIN)) {
            err = -EACCES;
            goto out;
        }
        err = unionfs_ioctl_superduper(inode, file, cmd, arg);
	break;

    default:
        if (ftopd(file) != NULL) {
	    hidden_file = ftohf(file);
	}
	/* pass operation to hidden filesystem, and return status */
	if (hidden_file && hidden_file->f_op && hidden_file->f_op->ioctl)
	    err = hidden_file->f_op->ioctl(itohi(inode), hidden_file, cmd, arg);
	else
	    err = -ENOTTY;	/* this is an unknown ioctl */
    } /* end of outer switch statement */

out:
    print_exit_status(err);
    return err;
}


/* FIST-LITE special version of mmap */
STATIC int
unionfs_mmap(file_t *file, vm_area_t *vma)
{
    int err = 0;
    file_t *hidden_file = NULL;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file); 

    PASSERT(hidden_file);
    PASSERT(hidden_file->f_op);
    PASSERT(hidden_file->f_op->mmap);

    vma->vm_file = hidden_file;
    err = hidden_file->f_op->mmap(hidden_file, vma);
    get_file(hidden_file); /* make sure it doesn't get freed on us */
    fput(file);			/* no need to keep extra ref on ours */

 out:
    print_exit_status(err);
    return err;
}


STATIC int
unionfs_open(inode_t *inode, file_t *file)
{
    int err = 0;
    int hidden_flags;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry = NULL;
    dentry_t *dentry = NULL;
    super_block_t *sb;
    int bindex = 0, bstart = 0, bend = 0;
    
    print_entry_location();

    ftopd(file) = KMALLOC(sizeof(struct unionfs_file_info), GFP_KERNEL);
    if (!ftopd(file)) {
	err = -ENOMEM;
	goto out;
    }
    init_ftopd(file);
    ftopd(file)->hash = KMALLOC(sizeof(struct unionfs_dir_hash), GFP_KERNEL);
    if (!ftopd(file)->hash) {
        err = -ENOMEM;
        goto out;
    }
    init_hash(file);
    atomic_set(&ftopd(file)->ufi_generation, atomic_read(&itopd(inode)->uii_generation));
    ftohf_ptr(file) = KMALLOC((sizeof(file_t *) * sbmax(inode->i_sb)), GFP_KERNEL);
    if (!ftohf_ptr(file)) {
        err = -ENOMEM;
        goto out;
    }
    init_ftohf_ptr(file, sbmax(inode->i_sb));
    
    hidden_flags = file->f_flags;
    
    dentry = file->f_dentry;
    PASSERT(dentry);

    bstart = fbstart(file) = dbstart(dentry);
    bend = fbend(file) = dbend(dentry);
 
    sb = inode->i_sb;
    
    /* open all directories and make the unionfs file struct point to these hidden file structs */
    if (S_ISDIR(inode->i_mode)) {

	for (bindex = bstart; bindex <= bend; bindex++) {
	    
            hidden_dentry = dtohd_index(dentry, bindex);
	    if (!hidden_dentry) {
		continue;
	    }
	    
	    dget(hidden_dentry);
	    /*
	     * dentry_open will decrement mnt refcnt if err.
	     * otherwise fput() will do an mntput() for us upon file close.
	     */
	    mntget(stohiddenmnt_index(sb, bindex));
	    branchget(sb, bindex);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), hidden_flags);
	    if (IS_ERR(hidden_file)) {
		err =  PTR_ERR(hidden_file);
		goto out;
	    }
	    
	    ftohf_index(file, bindex) = hidden_file;	/* link two files */
	}
    } else {
	/* open a file */
	hidden_dentry = dtohd(dentry);
	PASSERT(hidden_dentry);

        /* check for the permission for hidden file.  If the error is COPYUP_ERR, 
         * copyup the file.
         */
        if (hidden_dentry->d_inode) {
            /* if this is file is on RO branch, copyup the file */
	    if (IS_WRITE_FLAG(hidden_flags) && is_robranch(dentry)) {
		int size = file->f_dentry->d_inode->i_size;
		if (hidden_flags & O_TRUNC) {
			size = 0;
		}
                /* copyup the file */
                for (bindex = bstart - 1; bindex >= 0; bindex--) {
		    err = unionfs_copyup_file(dentry->d_parent->d_inode, file, bstart, bindex, size);
		    if (!err) {
                        break;
                    }
                }
                goto out;
            }
        }
	
        dget(hidden_dentry);
	/* dentry_open will decrement mnt refcnt if err.
         * otherwise fput() will do an mntput() for us upon file close.
	 */
	mntget(stohiddenmnt_index(sb, bstart));
	branchget(sb, bstart);
	hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bstart), hidden_flags);
	if (IS_ERR(hidden_file)) {
	    err =  PTR_ERR(hidden_file);
	    goto out;
	} else {
	    ftohf(file) = hidden_file;	/* link two files */
	}
    }

out:
    /* freeing the allocated resources, and fput the opened files */
    if (err < 0 && ftopd(file)) {
	for (bindex = bstart; bindex <= bend; bindex++) {
	    hidden_file  = ftohf_index(file, bindex);
	    if (hidden_file) {
		branchput(file->f_dentry->d_sb, bindex);	
		/* fput calls dput for hidden_dentry */
		fput(hidden_file);  
	    }
	}
	if (ftohf_ptr(file)) {
	    KFREE(ftohf_ptr(file));
	}
	KFREE(ftopd(file));
    }
    
    fist_print_file("OUT: unionfs_open", file);
    print_exit_status(err);
    return err;
}


STATIC int
unionfs_flush(file_t *file)
{
    int err = 0;		/* assume ok (see open.c:close_fp) */
    file_t *hidden_file = NULL;
    int bindex, bstart, bend;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    bstart = fbstart(file);
    bend = fbend(file);
    for (bindex = bstart; bindex <= bend; bindex++) {
	hidden_file  = ftohf_index(file, bindex);
	if (hidden_file && hidden_file->f_op && hidden_file->f_op->flush) {
	    err = hidden_file->f_op->flush(hidden_file);
	    if (err) {
		goto out;
	    }
	}
    }


out:
    print_exit_status(err);
    return err;
}


STATIC int
unionfs_release(inode_t *inode, file_t *file)
{
    int err = 0;
    file_t *hidden_file = NULL;
    int bindex, bstart, bend;

    print_entry_location();

    fist_checkinode(inode, "unionfs_release");

    /* fput all the hidden files */
    bstart = fbstart(file);
    bend = fbend(file);

    for (bindex = bstart; bindex <= bend; bindex++) {
      hidden_file = ftohf_index(file, bindex);

      if (hidden_file) {
        /*
         * will decrement file refcount, and if 0, destroy the file,
         * which will call the lower file system's file release function.
         */
        fput(hidden_file);
      }

      branchput(inode->i_sb, bindex);
    }
    KFREE(ftohf_ptr(file));

    if (ftopd(file)->hash) {
    	free_hash(ftopd(file)->hash);
    }
    KFREE(ftopd(file));

    fist_checkinode(inode, "post unionfs_release");

    print_exit_status(err);
    return err;
}


STATIC int
unionfs_fsync(file_t *file, dentry_t *dentry, int datasync)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    hidden_dentry = unionfs_hidden_dentry(dentry);

    if (hidden_file->f_op && hidden_file->f_op->fsync) {
	down(&hidden_dentry->d_inode->i_sem);
	err = hidden_file->f_op->fsync(hidden_file, hidden_dentry, datasync);
	up(&hidden_dentry->d_inode->i_sem);
    }

out:
    print_exit_status(err);
    return err;
}


STATIC int
unionfs_fasync(int fd, file_t *file, int flag)
{
    int err = 0;
    file_t *hidden_file = NULL;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);

    if (hidden_file->f_op && hidden_file->f_op->fasync)
	err = hidden_file->f_op->fasync(fd, hidden_file, flag);

out:
    print_exit_status(err);
    return err;
}


STATIC int
unionfs_lock(file_t *file, int cmd, struct file_lock *fl)
{
    int err = 0;
    file_t *hidden_file = NULL;

    print_entry_location();

    if ((err = unionfs_file_revalidate(file))) {
	goto out;
    }

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);

    if (hidden_file->f_op->lock) {
	err = hidden_file->f_op->lock(hidden_file, F_GETLK, fl);
    } else {
	posix_test_lock(hidden_file, fl);
    }

out:
    print_exit_status(err);
    return err;
}

/* Trimmed directory options, we shouldn't pass everything down since
 * we don't want to operate on partial directories.
 */
struct file_operations unionfs_dir_fops =
{
    read:	generic_read_dir,
    readdir:	unionfs_readdir,
    ioctl:	unionfs_ioctl,
    open:	unionfs_open,
    release:	unionfs_release,
    flush:	unionfs_flush,
};

struct file_operations unionfs_main_fops =
{
    llseek:	unionfs_llseek,
    read:	unionfs_read,
    write:	unionfs_write,
    readdir:	unionfs_file_readdir,
    poll:	unionfs_poll,
    ioctl:	unionfs_ioctl,
    mmap:	unionfs_mmap,
    open:	unionfs_open,
    flush:	unionfs_flush,
    release:	unionfs_release,
    fsync:	unionfs_fsync,
    fasync:	unionfs_fasync,
    lock:	unionfs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 * vim:shiftwidth=4
 */
