/*
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Charles P. Wright
 * Copyright (c) 2003-2004 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2004 Puja Gupta
 * Copyright (c) 2003-2004 Harikesavan Krishnan
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: subr.c,v 1.88 2004/12/29 15:11:47 cwright Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"

/* Determine the mode based on the copyup flags, and the existing dentry. */
static int unionfs_copyup_permissions(struct super_block *sb, struct dentry *old_hidden_dentry, struct dentry *new_hidden_dentry) {
    struct iattr newattrs;
    int err;

    newattrs.ia_valid = ATTR_CTIME;
    if (IS_SET(sb, COPYUP_FS_MOUNTER)) {
	/* f/s mounter */
	newattrs.ia_mode = stopd(sb)->copyupmode;
	newattrs.ia_valid |= ATTR_MODE;
	if (stopd(sb)->copyupuid != -1) {
	    newattrs.ia_uid = stopd(sb)->copyupuid;
	    newattrs.ia_valid |= ATTR_UID;
	}
	if (stopd(sb)->copyupgid != -1) {
	    newattrs.ia_gid = stopd(sb)->copyupgid;
	    newattrs.ia_valid |= ATTR_GID;
	}
    } else if (IS_SET(sb, COPYUP_CURRENT_USER)) {
	/* current file permission */
	newattrs.ia_mode = ~current->fs->umask & S_IRWXUGO;
	newattrs.ia_valid |= ATTR_MODE;
    } else {
	/* original mode of old file */
	newattrs.ia_mode = old_hidden_dentry->d_inode->i_mode;
	newattrs.ia_valid |= ATTR_MODE;
    }

    err = notify_change(new_hidden_dentry, &newattrs);
    return err;
}

STATIC int 
unionfs_copyup_dentry_len(inode_t *dir, dentry_t *dentry, int bstart, int new_bindex, file_t **copyup_file, int len)
{
    dentry_t *new_hidden_dentry;
    dentry_t *old_hidden_dentry = NULL;
    super_block_t *sb;
    file_t *input_file = NULL;
    file_t *output_file = NULL;
    ssize_t read_bytes, write_bytes;
    mm_segment_t old_fs;
    int err = 0;
    char *buf;
    int old_bindex;
    int got_branch_input = -1;
    int got_branch_output = -1;
    int old_bstart;
    int old_bend;
    int size = len;
    struct dentry *new_hidden_parent_dentry;
    mm_segment_t oldfs;
    char *symbuf = NULL;

        
    print_entry_location();
    
    fist_print_dentry("IN: unionfs_copyup_dentry_len: ", dentry);
        
    dget(dentry); 
    old_bindex = bstart;
    old_bstart = dbstart(dentry);
    old_bend = dbend(dentry);
    
    ASSERT(new_bindex >= 0);
    ASSERT(new_bindex < old_bindex);
    PASSERT(dir);
    PASSERT(dentry);
    
    sb = dir->i_sb;

    if((err = is_robranch_super(sb, new_bindex)))
	goto out;
   
    /* Create the directory structure above this dentry. */
    new_hidden_dentry = unionfs_create_dirs(dir, dentry, new_bindex);
    PASSERT(new_hidden_dentry);
    if (IS_ERR(new_hidden_dentry)) {
	err = PTR_ERR(new_hidden_dentry);
	goto out;
    }

    /* Now we actually create the object. */
    old_hidden_dentry = dtohd_index(dentry, old_bindex);
    PASSERT(old_hidden_dentry);
    PASSERT(old_hidden_dentry->d_inode);
    dget(old_hidden_dentry);

    /* For symlinks only, we have to read the link before we lock the directory. */
    if (S_ISLNK(old_hidden_dentry->d_inode->i_mode)) {
	PASSERT(old_hidden_dentry->d_inode->i_op);
	PASSERT(old_hidden_dentry->d_inode->i_op->readlink);

	symbuf = KMALLOC(PATH_MAX, GFP_KERNEL);
	if (!symbuf) {
	    err = -ENOMEM;
	    goto copyup_readlink_err;
	}

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	err = old_hidden_dentry->d_inode->i_op->readlink(old_hidden_dentry, symbuf, PAGE_SIZE);
	set_fs(oldfs);
	goto copyup_readlink_err;
    }

    /* Now we lock the parent, and create the object in the new branch. */
    new_hidden_parent_dentry = lock_parent(new_hidden_dentry);
    if (S_ISDIR(old_hidden_dentry->d_inode->i_mode)) {
	err = vfs_mkdir(new_hidden_parent_dentry->d_inode, new_hidden_dentry, S_IRWXU);
    } else if (S_ISLNK(old_hidden_dentry->d_inode->i_mode)) {
	err = vfs_symlink(new_hidden_parent_dentry->d_inode, new_hidden_dentry, symbuf);
    } else if (S_ISBLK(old_hidden_dentry->d_inode->i_mode) || S_ISCHR(old_hidden_dentry->d_inode->i_mode) || S_ISFIFO(old_hidden_dentry->d_inode->i_mode) || S_ISSOCK(old_hidden_dentry->d_inode->i_mode)) {
	err = vfs_mknod(new_hidden_parent_dentry->d_inode, new_hidden_dentry, old_hidden_dentry->d_inode->i_mode, kdev_t_to_nr(old_hidden_dentry->d_inode->i_rdev));
    } else if (S_ISREG(old_hidden_dentry->d_inode->i_mode)) {
	err = vfs_create(new_hidden_parent_dentry->d_inode, new_hidden_dentry, S_IRWXU);
    } else {
	FISTBUG("Unknown type of inode: %d\n", old_hidden_dentry->d_inode->i_mode);
    }
    unlock_dir(new_hidden_parent_dentry);
copyup_readlink_err:
    if (symbuf) {
	KFREE(symbuf);
    }
    if (err) {
   	/* get rid of the hidden dentry and all its traces */
    	dput(new_hidden_dentry);
	set_dtohd_index(dentry, new_bindex, NULL);
       	set_dbstart(dentry, old_bstart);
       	set_dbend(dentry, old_bend);
       	goto out;
    }

    /* We actually copyup the file here. */
    if (S_ISREG(old_hidden_dentry->d_inode->i_mode)) {
	mntget(stohiddenmnt_index(sb, old_bindex));
	branchget(sb, old_bindex);
	got_branch_input = old_bindex;
	input_file = dentry_open(old_hidden_dentry, stohiddenmnt_index(sb, old_bindex), O_RDONLY);
	if (IS_ERR(input_file)) {
	    err = PTR_ERR(input_file);
	    goto out;
	}
	if (!input_file->f_op || !input_file->f_op->read) {
	    err = -EINVAL;
	    goto out;
	}

	/* copy the new file */
	dget(new_hidden_dentry);
	mntget(stohiddenmnt_index(sb, new_bindex));
	branchget(sb, new_bindex);
	got_branch_output = new_bindex;
	output_file = dentry_open(new_hidden_dentry, stohiddenmnt_index(sb, new_bindex), O_WRONLY);
	if (IS_ERR(output_file)) {
	    err = PTR_ERR(output_file);
	    goto out;
	}
	if (!output_file->f_op || !output_file->f_op->write) {
	    err = -EINVAL;
	    goto out;
	}

	/* allocating a buffer */
	buf = (char *) KMALLOC(PAGE_SIZE, GFP_KERNEL);
	if (!buf) {
	    err = -ENOMEM;
	    goto out;
	}

	/* now read PAGE_SIZE bytes from offset 0 in a loop*/
	old_fs = get_fs();

	input_file->f_pos = 0;         
	output_file->f_pos = 0;

	set_fs(KERNEL_DS);
	do {
	    if (len >= PAGE_SIZE) {
		size = PAGE_SIZE;
	    } else if ((len < PAGE_SIZE) && (len > 0)) {
		size = len;
	    } 

	    len -= PAGE_SIZE;

	    read_bytes = input_file->f_op->read(input_file, buf, size, &input_file->f_pos);
	    if (read_bytes <= 0) {
		err = read_bytes;
		break;
	    }

	    write_bytes = output_file->f_op->write(output_file, buf, read_bytes, &output_file->f_pos);
	    if (write_bytes < 0 || (write_bytes < read_bytes)) {
		err = -EIO;
		break;
	    }
	} while ((read_bytes > 0) && (len > 0));
	set_fs(old_fs);
	KFREE(buf);
    }

    /* Set permissions. */
    if ((err = unionfs_copyup_permissions(sb, new_hidden_dentry, old_hidden_dentry))) {
	goto out;
    }

    unionfs_reinterpose(dentry);
   
 out:
    if (input_file && !IS_ERR(input_file)) {
    	fput(input_file);
    } else {
        /* since input file was not opened, we need to explicitly
         * dput the old_hidden_dentry
         */
        if (old_hidden_dentry) {
	   dput(old_hidden_dentry);
	}
    }

    /* in any case, we have to branchput */
    if (got_branch_input >= 0) {
        branchput(sb, got_branch_input);
    }

    if (output_file) {
    	if (copyup_file && !err) {
    	    *copyup_file = output_file;
    	} else {
            fput(output_file);
            branchput(sb, got_branch_output);
	}
    }

    dput(dentry);
    fist_print_dentry("OUT: unionfs_copyup_dentry_len: ", dentry);
    fist_print_inode("OUT: unionfs_copyup_dentry_len: ", dentry->d_inode);

    print_exit_status(err);
    return err;
}

/* This function creates a copy of a file represented by 'file' which currently 
 * resides in branch 'bstart' to branch 'new_bindex. 
 */
STATIC int
unionfs_copyup_file(inode_t *dir, file_t *file, int bstart, int new_bindex, int len)
{
    int err = 0;
    file_t *output_file = NULL;

    print_entry_location();

    err = unionfs_copyup_dentry_len(dir, file->f_dentry, bstart, new_bindex, &output_file, len);
    if (!err) {
        fbstart(file) = new_bindex;
        ftohf_index(file, new_bindex) = output_file;    
    }

out:
    print_exit_status(err);
    return err;
}


/* This function replicates the directory structure upto given dentry
 * in the bindex branch. Can create directory structure recursively to the right
 * also.
 */
STATIC dentry_t *
unionfs_create_dirs(inode_t *dir, dentry_t *dentry, int bindex)
{
    int err;
    dentry_t *child_dentry;
    dentry_t *parent_dentry;
    dentry_t *hidden_parent_dentry = NULL;
    dentry_t *hidden_dentry = NULL;

    const char *name;
    unsigned int namelen;

    dentry_t *path[MAX_DIR_CREATE];
    int count;          

    int old_bstart;
    int old_bend;

    print_entry_location();
  
    if ((err = is_robranch_super(dir->i_sb, bindex))) {
        hidden_dentry = ERR_PTR(err);
	goto out;
    }

    fist_print_dentry("IN: unionfs_create_dirs: ", dentry);
   
    old_bstart = dbstart(dentry);
    old_bend = dbend(dentry);

    memset(path, 0, sizeof(dentry_t *) * MAX_DIR_CREATE);

    /* assume the negative dentry of unionfs as the parent dentry */
    parent_dentry = dentry;
   
    count = 0;
    /* find out in the current branch the parent directory from 
     * where we need to start building a new directory structure
     */
    do {
        child_dentry = parent_dentry;
        
        /* find the parent directory dentry in unionfs */
        parent_dentry = child_dentry->d_parent;

        /* find out the hidden_parent_dentry in the given branch */
        hidden_parent_dentry = dtohd_index(parent_dentry, bindex);

        /* store the child dentry */
        path[count++] = child_dentry;
        
    } while(!hidden_parent_dentry);
    count--;

    while (1) {
        PASSERT(child_dentry);
        PASSERT(parent_dentry);
        PASSERT(parent_dentry->d_inode);
        
        // get hidden parent dir in the current branch
        hidden_parent_dentry = dtohd_index(parent_dentry, bindex);
        PASSERT(hidden_parent_dentry);
        PASSERT(hidden_parent_dentry->d_inode);

        // init the values to lookup
        name = child_dentry->d_name.name;
        namelen = child_dentry->d_name.len;
       
        // lookup child in the underlying file system
        hidden_dentry = lookup_one_len(name, hidden_parent_dentry, namelen);
        if (!hidden_dentry || IS_ERR(hidden_dentry)) {
            printk("ERR from hidden_dentry!!!\n");
            goto out;
        }

        if (child_dentry == dentry) {
            int loop_start;
            int loop_end;
            int new_bstart = -1;
            int new_bend = -1;
            int i;

            /* initialize the new dentry */
            set_dtohd_index(dentry, bindex, hidden_dentry);
            
            loop_start = (old_bstart < bindex) ? old_bstart : bindex;
            loop_end = (old_bend > bindex) ? old_bend : bindex;

            /* This loop sets the bstart and bend for the new
             * dentry by traversing from left to right.
             * It also dputs all negative dentries except 
             * bindex (the newly looked dentry
             */
            for ( i = loop_start; i <= loop_end; i++) {
                   
                if (!dtohd_index(dentry, i)) {
                    continue;
                }
                
                if ( i == bindex) {
                    new_bend = i;
                    if (new_bstart < 0) {
                        new_bstart = i;
                    }
                    continue;
                }

                if (!dtohd_index(dentry, i)->d_inode) {
                    dput(dtohd_index(dentry, i));
                    set_dtohd_index(dentry, i, NULL);
                } else {
                    if (new_bstart < 0) {
                        new_bstart = i;
                    }
                    new_bend = i;
                }
            }

            if (new_bstart < 0) {
                new_bstart = bindex;
            }
            if (new_bend < 0) {
                new_bend = bindex;
            }
            set_dbstart(dentry, new_bstart);
            set_dbend(dentry, new_bend);
            break;
        }

        if (hidden_dentry->d_inode) {
            /* since this already exists we dput to avoid 
             * multiple references on the same dentry
             */
            dput(hidden_dentry);
        } else {
	    mode_t newmode = S_IRWXU;

	    if (child_dentry->d_inode) {
		newmode = child_dentry->d_inode->i_mode & S_IRWXUGO;
	    }

            /* its a negative dentry, create a new dir */
            hidden_parent_dentry = lock_parent(hidden_dentry);
            err = vfs_mkdir(hidden_parent_dentry->d_inode, hidden_dentry, newmode);
            unlock_dir(hidden_parent_dentry);
            if (err || !hidden_dentry->d_inode) {
		dput(hidden_dentry);
                hidden_dentry = ERR_PTR(err);
                goto out;
            }

            itohi_index(child_dentry->d_inode, bindex) = igrab(hidden_dentry->d_inode);
            if (ibstart(child_dentry->d_inode) > bindex) {
                ibstart(child_dentry->d_inode) = bindex;
            }
	    if (ibend(child_dentry->d_inode) < bindex) {
		ibend(child_dentry->d_inode) = bindex;
	    }

            set_dtohd_index(child_dentry, bindex,  hidden_dentry);
            if (dbstart(child_dentry) > bindex) {
                set_dbstart(child_dentry, bindex);
            } 
            if (dbend(child_dentry) < bindex) {
                set_dbend(child_dentry,  bindex);
	    }
        } 

        parent_dentry = child_dentry;
        child_dentry = path[--count];
    }
out:
    fist_print_dentry("OUT: unionfs_create_dirs: ", dentry);
    print_exit_pointer(hidden_dentry);
    return hidden_dentry;
}


/* lookup files/directories from index 'bstart' onwards, does not looksup bstart */
STATIC int 
unionfs_partial_lookup(dentry_t *dentry)
{
    int err = 0;
    int dentry_count = 0;
    int bstart, bindex;
    const char *name;
    unsigned int namelen;
    dentry_t *hidden_dentry;
    dentry_t *hidden_dir_dentry;
    int positive;
    
    print_entry_location();
    
    fist_print_dentry("IN: unionfs_partial_lookup: ", dentry);
    lock_dpd(dentry);
    bstart = dbstart(dentry);
    
    name = dentry->d_name.name;
    namelen = dentry->d_name.len;

    positive = dentry->d_inode ? 1 :  0;

    for (bindex = bstart + 1; bindex < sbmax(dentry->d_sb); bindex++) {
	/* Check if there is already something there. */
	hidden_dentry = dtohd_index(dentry, bindex);
	if (hidden_dentry) {
	    continue;
	}
	
	/* Find out hidden parent dentry */
	hidden_dir_dentry = dtohd_index(dentry->d_parent, bindex);
	
	/* continue if hidden parent doesn't exist */
	if (!hidden_dir_dentry) {
	    continue;
	}
	
	/* lookup underlying file system */
	hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);
	if (IS_ERR(hidden_dentry)) {
	    printk("ERR from hidden_dentry!!!\n");
	    err = PTR_ERR(hidden_dentry);
	    goto out;
	}
	
	if (hidden_dentry->d_inode) {
	    /* this is not a negative dentry, so init dtohd with 
	     * this new dentry.  Also, change bend.
	     */
	    set_dtohd_index(dentry, bindex, hidden_dentry);
	    set_dbend(dentry, bindex);
	    dentry_count++;
	} else {
	    /* dput the unnecessary negative dentry */
	    dput(hidden_dentry);
	}
    } // end for 
    
    if (dentry_count) {
	if (positive) {
    		/* need to connect the new inodes to the pre-existing 
     		 * unionfs inode, unionfs_reinterpose
      		 */
		unionfs_reinterpose(dentry);
	} else {
		int bstart;
		int bend;

		/* Somehow the file was created for us, so we need to interpose it with a new inode. */

		bstart = dbstart(dentry);
		bend = dbend(dentry);

        	for (bindex = bstart; bindex <= bend; bindex++) {
               		if (dtohd_index(dentry, bindex)) {
				if (!dtohd_index(dentry, bindex)->d_inode) {
					dput(dtohd_index(dentry, bindex));
					set_dtohd_index(dentry, bindex, NULL);
				} else {
					set_dbstart(dentry, bindex);
                        		break;
				}
			}
		}
		
		unionfs_interpose(dentry, dentry->d_sb, 0);
	}
    }
    
 out:
    unlock_dpd(dentry);
    fist_print_dentry("OUT: unionfs_partial_lookup: ", dentry);
    print_exit_status(err);
    return err;
}


struct create_dir_whs_callback {
    int err;
    dentry_t *dentry;
    int write_index;
    int filldir_called;
};


int vfs_create_nozombie(struct inode *dir, struct dentry *dentry, int mode);
/* This is designed to be called just like a filldir function. */
static int create_dir_whs_filldir(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type)
{
    int err = 0;
    struct create_dir_whs_callback *buf = (struct create_dir_whs_callback *) dirent;
    dentry_t *dentry;
    int write_index;
    dentry_t *hidden_parent_dentry;
    dentry_t *hidden_dentry;
    dentry_t *hidden_whiteout_dentry;
    char *wh_name = NULL;

    buf->filldir_called = 1;
    
    /* skip '.' and '..' entries */
    if (name[0] == '.' && (namelen == 1 || (name[1] == '.' && namelen == 2))) {
        goto out;
    }
   
    wh_name = KMALLOC(sizeof(char) * (namelen + 5), GFP_KERNEL);
    if (!wh_name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(wh_name, ".wh.");
    strncat(wh_name, name, namelen);
    wh_name[4 + namelen] = '\0';
   
    dentry = buf->dentry;
    write_index = buf->write_index;
    
    hidden_dentry = dtohd_index(dentry, write_index);
    hidden_whiteout_dentry = lookup_one_len(wh_name, hidden_dentry, namelen + 4);
    if (IS_ERR(hidden_dentry)) {
	/*
	 * this produces an unusual dentry: one that has neither an
	 * inode, nor a private structure attached to it. All cleanup
	 * methods (d_delete, d_release, etc) must be prepared to deal
	 * with such dentries. Ion 09/29/2001
	 */
	printk("ERR from hidden_dentry!!!\n");
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    if (!hidden_whiteout_dentry->d_inode) {
    	hidden_parent_dentry = lock_parent(hidden_whiteout_dentry);

    	err = vfs_create_nozombie(hidden_parent_dentry->d_inode, hidden_whiteout_dentry, 0666);
    	fist_copy_attr_times(dentry->d_inode, hidden_parent_dentry->d_inode);
    	unlock_dir(hidden_parent_dentry);
    }
    dput(hidden_whiteout_dentry);
    
 out:
    if (wh_name) {
        KFREE(wh_name);
    }
    buf->err = err;
    return err;
}

/* Creates a whiteout for the files existing in a directory.  Pass the unionfs dentry,
 * and the bindex
 */
int create_dir_whs(dentry_t *dentry, int cur_index) 
{
    int err = 0;
    int bindex = 0, bstart, bend;
    dentry_t *hidden_dentry = NULL;
    const char *name;
    unsigned int namelen;
    dentry_t *hidden_parent_dentry;
    super_block_t *sb;
    file_t *hidden_file;
    struct create_dir_whs_callback buf;

    sb = dentry->d_sb;

    name = dentry->d_name.name;
    namelen = dentry->d_name.len;
    
    bstart = dbstart(dentry);
    bend = dbend(dentry);
    
    print_entry_location();
    
    fist_print_dentry("IN create_dir_whs: ", dentry);
     
    /* open all dentries to the right of branch 'cur_index' and do vfs_readdir on them.
       In the create_dir_whs_filldir() function, for each entry create a whiteout in branch 'cur_index'  */
    for (bindex = cur_index + 1; bindex <= sbend(sb); bindex++) {
	
	/* Find out hidden parent dentry */
	hidden_parent_dentry = dtohd_index(dentry->d_parent, bindex);
	
	/* continue if hidden parent doesn't exist */
	if (!hidden_parent_dentry) {
	    continue;
	}
	
	/* lookup underlying file system */
	hidden_dentry = lookup_one_len(name, hidden_parent_dentry, namelen);
	if (IS_ERR(hidden_dentry)) {
	    err = PTR_ERR(hidden_dentry);
	    goto out;
	}
	
	if (hidden_dentry->d_inode && S_ISDIR(hidden_dentry->d_inode->i_mode)) {
	    
	    mntget(stohiddenmnt_index(sb, bindex));
	    branchget(sb, bindex);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), O_RDONLY);
	    if (IS_ERR(hidden_file)) {
		err = PTR_ERR(hidden_file);
                branchput(sb, bindex);
		goto out;
	    }
	    /* prepare callback buffer */
	    buf.dentry = dentry;
	    buf.write_index = cur_index;
	    
	    do {
	        buf.filldir_called = 0;
	        err = vfs_readdir(hidden_file, create_dir_whs_filldir, &buf);
	        if (buf.err) {
		    err = buf.err;
	        }
	    } while (!err && buf.filldir_called);

	    /* fput calls dput for internally */
            fput(hidden_file);  
            branchput(sb, bindex);
	    
	} else {
	    /* dput the unnecessary negative dentry */
	    dput(hidden_dentry);
	}
    }
    
 out:
    fist_print_dentry("OUT create_dir_whs :", dentry);
    print_exit_status(err);
    return err;
}

/* Pass an unionfs dentry and an index.  It will try to create a whiteout for the filename
 * in dentry, and will try in branch 'index'.  On error, it will proceed to a branch
 * to the left
 */
int create_whiteout(dentry_t *dentry, int start)
{
    int bstart, bend, bindex;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_dentry;
    dentry_t *hidden_wh_dentry;
    char *name = NULL;
    int err = -EUNIONFS_NO_WHITEOUT;

    print_entry_location();

    PASSERT(dentry);
    PASSERT(dentry->d_parent);

    fist_print_dentry("IN create_whiteout", dentry);
    bstart = dbstart(dentry);
    bend = dbend(dentry);
    
    /* create dentry's whiteout equivalent */
    name = KMALLOC(sizeof(char) * (dentry->d_name.len + 5), GFP_KERNEL);
    if (!name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(name, ".wh.");
    strncat(name, dentry->d_name.name, dentry->d_name.len);
    name[dentry->d_name.len + 4] = '\0';

    /* start from 'start', create foo then rename it to .wh.foo, on error go to 
       the next left branch, if hidden_dentry is NULL do recursive directory creation */
    for (bindex = start; bindex >= 0; bindex--) {
	hidden_dentry = dtohd_index(dentry, bindex);
	
	if (!hidden_dentry) {
	    /* if hidden dentry is not present, create the entire 
	     * hidden dentry directory structure and go ahead.
	     * Since we want to just create whiteout, we only want the parent
	     * dentry, and hence get rid of this dentry.
	     */
	    hidden_dentry = unionfs_create_dirs(dentry->d_inode, dentry, bindex);
	    if (!hidden_dentry || IS_ERR(hidden_dentry)) {
		fist_dprint(8, "hidden dentry NULL for bindex = %d\n", bindex);
		continue;
	    }
	}
	hidden_wh_dentry = lookup_one_len(name, hidden_dentry->d_parent, dentry->d_name.len + 4);

	if (!hidden_wh_dentry || IS_ERR(hidden_wh_dentry)) {
	    continue;
	}
	ASSERT(!hidden_wh_dentry->d_inode);
	
	hidden_dir_dentry = lock_parent(hidden_wh_dentry);
    	if (!(err = is_robranch_super(dentry->d_sb, bindex))) {
	    err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO);
	}
	unlock_dir(hidden_dir_dentry);
	dput(hidden_wh_dentry);
        

	if (!err) {
	    break;
        } else {
            if (IS_SET(dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                goto out;
            }
            /* break out of for loop if error returned was NOT -EROFS */
            if (!IS_COPYUP_ERR(err)) {
                break;
            }
        }
    } // end for

    fist_print_dentry("OUT create_whiteout", dentry);
out:
    if (name) {
       KFREE(name);
    }
    print_exit_status(err);
    return err;
}

/* Pass an unionfs dentry, name of file, and an index.  It will try to create a whiteout 
 * for the filename passed in the parent directory passed.  It will start trying 
 * in branch start.  On error, it will proceed to a branch to the left
 */
int create_whiteout_parent(dentry_t *parent_dentry, const char *filename, int start)
{
    int bindex;
    int old_bstart, old_bend;
    dentry_t *hidden_dir_dentry;
    dentry_t *hidden_grand_parent_dentry;
    dentry_t *hidden_parent_dentry;
    dentry_t *hidden_wh_dentry;
    char *name = NULL;
    int err = -EUNIONFS_NO_WHITEOUT;

    print_entry_location();

    PASSERT(parent_dentry);
    old_bstart = dbstart(parent_dentry);
    old_bend = dbend(parent_dentry);
 
    fist_print_dentry("IN create_whiteout_parent", parent_dentry);
     
    /* create dentry's whiteout equivalent */
    name = KMALLOC(sizeof(char) * (strlen(filename) + 5), GFP_KERNEL);
    if (!name) {
        err = -ENOMEM;
        goto out;
    }
    strcpy(name, ".wh.");
    strncat(name, filename, strlen(filename));
    name[strlen(filename) + 4] = '\0';

    for (bindex = start; bindex >= 0; bindex--) {
        
        hidden_parent_dentry = dtohd_index(parent_dentry, bindex);

	if (!hidden_parent_dentry) {
            /* create the recursive directory structure and return the 
             * negative dentry for the parent where we want to create whiteout
             */
            ASSERT(parent_dentry->d_inode != NULL);
	    hidden_parent_dentry = unionfs_create_dirs(parent_dentry->d_inode, parent_dentry, bindex);
	    if (!hidden_parent_dentry || IS_ERR(hidden_parent_dentry)) {
		fist_dprint(8, "hidden dentry NULL for bindex = %d\n", bindex);
		continue;
	    } else {
	        /* create directory of the hidden parent, if it is negative. 
                 * This is where whiteout is created 
                 */
                hidden_grand_parent_dentry = lock_parent(hidden_parent_dentry);
                PASSERT(hidden_grand_parent_dentry->d_inode);

                /* We shouldn't create things in a read-only branch. */
                if (!(err = is_robranch_super(parent_dentry->d_sb, bindex))) {
                    err = vfs_mkdir(hidden_grand_parent_dentry->d_inode, hidden_parent_dentry, S_IRWXU);
                }
                
                unlock_dir(hidden_grand_parent_dentry);
                if (err || !hidden_parent_dentry->d_inode) {
		    dput(hidden_parent_dentry);
                    /* if error, we need to revert back the old bstart and bend for this
                     * parent and return back
                     */
                    set_dbstart(parent_dentry, old_bstart);
                    set_dbend(parent_dentry, old_bend);
                    set_dtohd_index(parent_dentry, bindex, NULL);
                    
                    if (IS_SET(parent_dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                        goto out;
                    }
                    /* break out of for loop if error returned was NOT -EROFS */
                    if (!IS_COPYUP_ERR(err)) {
                        break;
                    }
                    
                    continue;
                }
                itohi_index(parent_dentry->d_inode, bindex) = igrab(hidden_parent_dentry->d_inode);
            }
        }
        
        /* lookup for the whiteout dentry that we want to create */
	hidden_wh_dentry = lookup_one_len(name, hidden_parent_dentry, strlen(filename) + 4);
	if (!hidden_wh_dentry || IS_ERR(hidden_wh_dentry)) {
	    continue;
	}
	ASSERT(!hidden_wh_dentry->d_inode);
	
        /* hidden_dir_dentry and hidden_parent_dentry are going to be same only */
	hidden_dir_dentry = lock_parent(hidden_wh_dentry);

        /* We shouldn't create things in a read-only branch. */
        if (!(err = is_robranch_super(parent_dentry->d_sb, bindex))) {
            err = vfs_create(hidden_dir_dentry->d_inode, hidden_wh_dentry, ~current->fs->umask & S_IRWXUGO);
	}
        
        unlock_dir(hidden_dir_dentry);
	dput(hidden_wh_dentry);

	if (!err) {
	    break;
        } else {
            if (IS_SET(parent_dentry->d_sb, GLOBAL_ERR_PASSUP)) {
                goto out;
            }
            /* break out of for loop if error returned was NOT -EROFS */
            if (!IS_COPYUP_ERR(err)) {
                break;
            }
        }
    } // end for

out:
    fist_print_dentry("OUT create_whiteout", parent_dentry);
    if (name) {
       KFREE(name);
    }
    print_exit_status(err);
    return err;
}

/* The callback structure for our whiteouts. */
struct unionfs_check_whiteout_callback {
    int err;
    struct list_head dir_list[HASHTABLE_SIZE];
    int filldir_called;
};

/* This is designed to be called just like a filldir function. */
static int check_whiteout(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type) 
{
    int err = 0;
    struct unionfs_check_whiteout_callback *buf = (struct unionfs_check_whiteout_callback *) dirent;
    int whiteout = 0;
    int bucket;
    struct list_head *head;
    struct filldir_node *cursor;
    struct filldir_node *new = NULL;
    int found = 0;
    unsigned int hash;

    print_entry_location();

    buf->filldir_called = 1;
          
    if (name[0] == '.' && (namelen == 1 || (name[1] == '.' && namelen == 2))) {
	goto out;
    }

    if (!strncmp(name, ".wh.", 4)) {
	namelen = namelen - 4;
	hash = full_name_hash(name + 4, namelen);
	name = name + 4;	
	whiteout = 1;
    } else {
	hash = full_name_hash(name, namelen);
    }

    bucket = hash % HASHTABLE_SIZE;

    head = &(buf->dir_list[bucket]);

    list_for_each_entry(cursor, head, file_list) {
	if (cursor->namelen == namelen && cursor->hash == hash && !strncmp(name, cursor->name, namelen)) {
	    found = 1;
	    break;
	}
    }

    /* If it was found in the table there was a previous whiteout. */
    if (found) {
	goto out;
    }

    /* If it wasn't found, and it isn't a whiteout the directory is not actually empty. */
    if (!whiteout) {
	err = -ENOTEMPTY;
	goto out;
    }    

    /* Now insert the whiteout entry into the table. */
    new = KMALLOC(sizeof(struct filldir_node), GFP_KERNEL);
    if (!new) {
        err = -ENOMEM;
        goto out;
    }
    INIT_LIST_HEAD(&new->file_list);
    new->namelen = namelen;
    new->hash = hash;
    new->bindex = -1; /* We don't keep track here. */
    new->name = (char *) KMALLOC(namelen, GFP_KERNEL);
    if (!new->name) {
	err = -ENOMEM;
	goto out;	
    } 
    memcpy(new->name, name, namelen);
    list_add(&(new->file_list), head);
    new = NULL;    

 out:
    if (new) {
	KFREE(new);
    }
    buf->err = err;
    print_exit_status(err);
    return err;
}


// used by rmdir to check if a directory is indeed empty.
int check_empty(dentry_t *dentry) 
{
    int err = 0;
    dentry_t *hidden_dentry = NULL;
    super_block_t *sb;
    file_t *hidden_file;
    struct unionfs_check_whiteout_callback buf;
    int bindex, bstart, bend;
    int i;
    struct list_head *head;
    struct list_head *pos;
    struct list_head *n;
    struct filldir_node *cursor;
    
    print_entry_location();

    sb = dentry->d_sb;

    buf.err = 0;
    for (i = 0; i < HASHTABLE_SIZE; i++) {
	INIT_LIST_HEAD(&buf.dir_list[i]);
    }

    ASSERT(S_ISDIR(dentry->d_inode->i_mode));

    /* First we need to make sure all of the hidden_dentries are filled in. */
    err = unionfs_partial_lookup(dentry);
    if (err) {
	goto out;
    }

    bstart = dbstart(dentry);
    bend = dbend(dentry);
    
    /* Open all of the hidden_dentries and use the check_whiteout function as a filldir. */
    for (bindex = bstart; bindex <= bend; bindex++) {
	/* Find out hidden parent dentry */
	hidden_dentry = dtohd_index(dentry, bindex);
	
	/* continue if the hidden_dentry doesn't exist */
	if (!hidden_dentry) {
	    continue;
	}

	/* If it is a positive dentry. */
	if (hidden_dentry->d_inode && S_ISDIR(hidden_dentry->d_inode->i_mode)) {
	    dget(hidden_dentry);
	    mntget(stohiddenmnt_index(sb, bindex));
	    branchget(sb, bindex);
	    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), O_RDONLY);
	    if (IS_ERR(hidden_file)) {
		err = PTR_ERR(hidden_file);
		dput(hidden_dentry);
                branchput(sb, bindex);
		goto out;
	    }
	   
	    do {
	        buf.filldir_called = 0;
		err = vfs_readdir(hidden_file, check_whiteout, &buf);
	        if (buf.err) {
		    err = buf.err;
	        }
	    } while (!err && buf.filldir_called);

	    /* fput calls dput for hidden_dentry */
	    fput(hidden_file); 
            branchput(sb, bindex);

	    if (err < 0) 
		goto out;
	}
    }

 out:
    for(i = 0; i < HASHTABLE_SIZE; i++) {
	head = &(buf.dir_list[i]);

	list_for_each_safe(pos, n, head) {
	    cursor = list_entry(pos, struct filldir_node, file_list);
            list_del(&cursor->file_list);
	    KFREE(cursor->name);
	    KFREE(cursor);
	}
    }
    
    print_exit_status(err);
    return err;
}

/* The callback structure for our whiteouts. */
struct unionfs_delete_whiteout_callback {
    int err;
    int bindex;
    struct dentry *parent;
    int filldir_called;
};

int vfs_unlink_nozombie(struct inode *dir, struct dentry *dentry);

/* This is designed to be called just like a filldir function. */
static int delete_whiteout_actor(void *dirent, const char *name, int namelen, loff_t offset, ino_t ino, unsigned int d_type) {
    int err = 0;
    struct unionfs_delete_whiteout_callback *buf = (struct unionfs_delete_whiteout_callback *) dirent;
    struct dentry *hidden_dentry;
    struct dentry *hidden_dir_dentry;

    print_entry_location();

    buf->filldir_called = 1;

    if (name[0] == '.' && (namelen == 1 || (name[1] == '.' && namelen == 2))) {
	goto out;
    }

    if (strncmp(name, ".wh.", 4)) {
	err = -ENOTEMPTY;
	goto out;
    }
    hidden_dir_dentry = dtohd_index(buf->parent, buf->bindex);
    PASSERT(hidden_dir_dentry);
    hidden_dentry = lookup_one_len(name, hidden_dir_dentry, namelen);
    if (IS_ERR(hidden_dentry)) {
	err = PTR_ERR(hidden_dentry);
	goto out;
    }	
    if (!hidden_dentry->d_inode) {
	dput(hidden_dentry);
        goto out;
    }

    /* we don't	use lock_parent because readdir already has us locked. */
    hidden_dir_dentry = hidden_dentry->d_parent;
	
    /* avoid destroying the hidden inode if the file is in use */
    err = vfs_unlink_nozombie(hidden_dir_dentry->d_inode, hidden_dentry);
    dput(hidden_dentry);
    
    if (!err) {			/* vfs_unlink does that */
    	d_delete(hidden_dentry); 
    } else {
        /* exit if the error returned was NOT -EROFS */
        if (!IS_COPYUP_ERR(err)) {
            goto out;
        }
    }
    fist_copy_attr_times(buf->parent->d_inode, hidden_dir_dentry->d_inode);
    /* propagate number of hard-links */
    buf->parent->d_inode->i_nlink = get_nlinks(buf->parent->d_inode);

 out:
    buf->err = err;
    print_exit_status(err);
    return err;
}

// used by rmdir to nuke whiteouts in a directory
int delete_whiteouts(dentry_t *dentry, int bindex) 
{
    int err = 0;
    dentry_t *hidden_dentry = NULL;
    super_block_t *sb;
    file_t *hidden_file = NULL;
    struct unionfs_delete_whiteout_callback buf;
    
    print_entry_location();

    sb = dentry->d_sb;

    buf.err = 0;  

    buf.bindex = bindex;
    buf.parent = dentry;

    ASSERT(S_ISDIR(dentry->d_inode->i_mode));

    /* First we need to make sure all of the hidden_dentries are filled in. */
    err = unionfs_partial_lookup(dentry);
    if (err) {
	goto out;
    }

    ASSERT(bindex >= dbstart(dentry));
    ASSERT(bindex <= dbend(dentry));

    /* Find out hidden parent dentry */
    hidden_dentry = dtohd_index(dentry, bindex);
    PASSERT(hidden_dentry);
    PASSERT(hidden_dentry->d_inode);
    ASSERT(S_ISDIR(hidden_dentry->d_inode->i_mode));

    dget(hidden_dentry);
    mntget(stohiddenmnt_index(sb, bindex));
    branchget(sb, bindex);
    hidden_file = dentry_open(hidden_dentry, stohiddenmnt_index(sb, bindex), O_RDONLY);
    if (IS_ERR(hidden_file)) {
	err = PTR_ERR(hidden_file);
	dput(hidden_dentry);
        branchput(sb, bindex);
	goto out;
    }
    
    do {
        buf.filldir_called = 0;
	err = vfs_readdir(hidden_file, delete_whiteout_actor, &buf);
        if (buf.err) {
            err = buf.err;
        }
    } while (!err && buf.filldir_called);
	
    /* fput calls dput on the hidden_dentry */
    fput(hidden_file);
    branchput(sb, bindex);

 out:
    print_exit_status(err);
    return err;
}

/* This is a helper function for rename, which ends up with hosed over dentries
 * when it needs to revert. */
int unionfs_refresh_hidden_dentry(struct dentry *dentry, int bindex) {
    struct dentry *hidden_dentry;
    struct dentry *hidden_parent;
    int err = 0;
	
    print_entry(" bindex = %d", bindex);

    hidden_parent = dtohd_index(dentry->d_parent, bindex);

    PASSERT(hidden_parent);
    PASSERT(hidden_parent->d_inode);
    ASSERT(S_ISDIR(hidden_parent->d_inode->i_mode));

    hidden_dentry = lookup_one_len(dentry->d_name.name, hidden_parent, dentry->d_name.len);
    if (IS_ERR(hidden_dentry)) {
	err = PTR_ERR(hidden_dentry);
	goto out;
    }

    if (dtohd_index(dentry, bindex)) {
    	dput(dtohd_index(dentry, bindex));
    }
    if(itohi_index(dentry->d_inode, bindex)) {
    	iput(itohi_index(dentry->d_inode, bindex));
	itohi_index(dentry->d_inode, bindex) =  NULL;
    }
    if (!hidden_dentry->d_inode) {
	dput(hidden_dentry);
    	set_dtohd_index(dentry, bindex, NULL);
    } else {
    	set_dtohd_index(dentry, bindex, hidden_dentry);
    	itohi_index(dentry->d_inode, bindex) = igrab(hidden_dentry->d_inode);
    }

out:
    print_exit_status(err);
    return err;
}

/*
 * vim:shiftwidth=4
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
